"""
Primary entry point for schemate CLI application
"""

import os
import sys
import colorama
import argparse

from .config import Config
from .version import get_version
from .exceptions import SchemateException, CommandError


def analyze(args: argparse.Namespace) -> None:
    """
    CLI entry point for analyzing documents.
    """
    config = Config.load(args.config)
    print(config)


##########################################################################
## CLI Helpers
##########################################################################

class SchemateFormatter(argparse.ArgumentDefaultsHelpFormatter):

    def __init__(self, *args, **kwargs):
        if "width" not in kwargs:
            try:
                kwargs["width"] = os.get_terminal_size().columns
            except OSError:
                pass

        if "max_help_position" not in kwargs:
            kwargs["max_help_position"] = 32

        super().__init__(*args, **kwargs)


class Environ(argparse.Action):

    def __init__(self, envvar, required=True, default=None, **kwargs):
        if envvar:
            if envvar in os.environ:
                default = os.environ[envvar]
        if required and default:
            required = False
        super(Environ, self).__init__(default=default, required=required, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, values)


##########################################################################
## CLI Main Entry Point
##########################################################################

# argparse application description
DESCRIPTION = "Analyzes groups of documents and generates a schema profile for them."
EPILOG = "Created with ♠ by the Rotational Labs team."

# argparse commands
CMDS = {
    "analyze": {
        "help": "analyze documents and generate a schema profile",
        "func": analyze,
        "args": {
            ("-c", "--config"): {
                "metavar": "PATH", "default": None, "type": str, ""
                "action": Environ, "envvar": "SCHEMATE_CONFIG",
                "help": "path to the configuration file for the analysis (env: $SCHEMATE_CONFIG)",
            },
            ("-o", "--output"): {
                "metavar": "PATH", "default": None, "type": str,
                "help": "write the schema profile to a file instead of stdout",
            },
        },
    },
}


def main():
    """
    Main entry point for schemate CLI application.
    """
    colorama.init(autoreset=True)
    parser = argparse.ArgumentParser(
        prog="schemate",
        description=DESCRIPTION,
        epilog=EPILOG,
        formatter_class=SchemateFormatter,
    )

    # Add version information
    parser.add_argument(
        "-v", "--version",
        action="version",
        version=f"%(prog)s {get_version()}",
        help="Show the version and exit.",
    )

    # Create subparsers for each command
    subparsers = parser.add_subparsers(title="actions")
    for cmd, cargs in CMDS.items():
        ckws = {"help": cargs.get("help"), "formatter_class": SchemateFormatter}
        subparser = subparsers.add_parser(cmd, **ckws)
        subparser.set_defaults(func=cargs.get("func"))

        for pargs, kwargs in cargs.get("args", {}).items():
            if isinstance(pargs, str):
                pargs = (pargs,)
            subparser.add_argument(*pargs, **kwargs)

    # Execute the argparse CLI parser and associated command
    args = parser.parse_args()
    if hasattr(args, "func"):
        try:
            args.func(args)
        except CommandError as e:
            parser.error(e)
        except SchemateException as e:
            print(colorama.Fore.RED + str(e))
            sys.exit(2)
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



"""
Data loaders for iterating through schemaless datasources
"""

import os
import glob
import json
import yaml

from pymongo import MongoClient
from collections.abc import Iterable, Sized
from .exceptions import UnsupportedLoader, LoaderError


class Loader(Iterable, Sized):
    """
    A loader class is an interator that reads schemaless datasources such as a MongoDB
    collection or a directory of files and yields Python dictionaries with the
    deserialized data from those sources. It may also provide other methods for
    connection management and filtering.
    """

    def count(self):
        """
        Returns the total number of documents being managed by the loader.
        """
        return len(self)


class FileLoader(Loader):
    """
    Loads documents from a single yaml, json, or jsonlines file based on the extension.
    If the file is yaml or json, it yields a single document (count=1). If it is a
    jsonlines file it yields one document per line (count=n).

    Parameters
    ----------
    path : str
        The path to the file to load. Its extension must be one of the supported types.
    """

    SUPPORTED_TYPES = {
        ".json", ".jsonlines", ".jsonl", ".yaml", ".yml"
    }

    @staticmethod
    def is_supported(path: str) -> bool:
        """
        Returns true if the extension of the specified path is one of the supported
        file types that can be loaded by this class.
        """
        return FileLoader.extension(path) in FileLoader.SUPPORTED_TYPES

    @staticmethod
    def extension(path: str) -> str:
        """
        Returns the normalized extension of the specified path.
        """
        return os.path.splitext(os.path.basename(path))[1].lower()

    def __init__(self, path: str):
        self._path = path
        self._ext = self.extension(path)
        self._count = None

        if self._ext not in self.SUPPORTED_TYPES:
            raise UnsupportedLoader(f"unsupported file type: {self._ext}")

    def __len__(self):
        if self._count is None:
            # In this case we haven't performed an iteration, so count manually.
            if self._ext in {".jsonlines", ".jsonl"}:
                self._count = sum(1 for _ in self.load_jsonlines())
            else:
                self._count = 1
        return self._count

    def __iter__(self):
        self._count = 0
        for document in self.load():
            self._count += 1
            yield document

    def load(self):
        """
        Uses the extension of the wrapped path to return the data in the file.
        """
        try:
            if self._ext == ".json":
                return self.load_json()
            elif self._ext in {".yaml", ".yml"}:
                return self.load_yaml()
            elif self._ext in {".jsonlines", ".jsonl"}:
                return self.load_jsonlines()
            else:
                raise UnsupportedLoader(f"unsupported file type: {self._ext}")
        except OSError as e:
            raise LoaderError(f"unable to open file: {self._path}") from e
        except FileNotFoundError as e:
            raise LoaderError(f"missing data file: {self._path}") from e

    def load_json(self) -> dict:
        """
        Loads the file as a single JSON document.
        """
        with open(self._path, "r") as f:
            return [json.load(f)]

    def load_yaml(self) -> dict:
        """
        Loads the file as a single YAML document.
        """
        with open(self._path, "r") as f:
            return [yaml.safe_load(f)]

    def load_jsonlines(self) -> Iterable[dict]:
        """
        Loads a jsonlines file where each line is a JSON document.
        """
        with open(self._path, "r") as f:
            for line in f:
                yield json.loads(line)


class MultiFileLoader(Loader):
    """
    Loads documents from a list of files. The files can be of any supported type by the
    FileLoader class. The loader will iterate through the files and yield documents from
    each file using the FileLoader.

    Parameters
    ----------
    paths : list[str]
        A list of paths to the files to load.

    ignore_unsupported : bool (default: True)
        If True, the paths will be filtered to exclude unsupported file types; otherwise
        an exception will be raised if any unsupported file types are found.
    """

    def __init__(self, paths: list[str], ignore_unsupported: bool = True):
        if ignore_unsupported:
            paths = [
                path for path in paths if FileLoader.is_supported(path)
            ]

        self._count = None
        self._loaders = [
            FileLoader(path) for path in paths
        ]

    def __len__(self):
        if self._count is None:
            self._count = sum(len(loader) for loader in self._loaders)
        return self._count

    def __iter__(self):
        self._count = 0
        for loader in self._loaders:
            for document in loader.load():
                self._count += 1
                yield document

    def filenames(self):
        for loader in self._loaders:
            yield loader._path


class DirectoryLoader(MultiFileLoader):
    """
    Loads all the documents in a directory and optionally all subdirectories. The
    directories must only contain documents of the supported types in the FileLoader
    class (otherwise those files will be ignored). This class also supports loading of
    multiple directories.

    Parameters
    ----------
    dirs : str | list[str]
        A directory path or a list of directory paths to load files from.

    recursive : bool (default: False)
        If True, the loader will recursively load all files in the directory and its
        subdirectories. Otherwise, it will only load files in the top-level directory.

    ignore_unsupported : bool (default: True)
        If True, the paths will be filtered to exclude unsupported file types; otherwise
        an exception will be raised if any unsupported file types are found.
    """

    def __init__(self, dirs: str | list[str], recursive: bool = False, ignore_unsupported: bool = True):
        paths = []
        if isinstance(dirs, str):
            dirs = [dirs]

        # Get all the files in the directories and subdirectories if recursive.
        for dir in dirs:
            for root, _, filenames in os.walk(dir):
                for filename in filenames:
                    paths.append(os.path.join(root, filename))
                if not recursive:
                    break

        super().__init__(paths, ignore_unsupported=ignore_unsupported)


class GLOBLoader(MultiFileLoader):
    """
    Loads all files that match the specified glob pattern(s).

    Parameters
    ----------
    patterns : str | list[str]
        A glob pattern or a list of glob patterns to load files from.

    ignore_unsupported : bool (default: True)
        If True, the paths will be filtered to exclude unsupported file types; otherwise
        an exception will be raised if any unsupported file types are found.
    """

    def __init__(self, patterns: str | list[str], ignore_unsupported: bool = True):
        paths = []
        if isinstance(patterns, str):
            patterns = [patterns]

        for pattern in patterns:
            paths.extend(glob.glob(pattern))

        super().__init__(paths, ignore_unsupported=ignore_unsupported)


class MongoDBLoader(Loader):
    """
    Loads documents from a collection (or collections) in a MongoDB database.

    Parameters
    ----------
    uri : str
        The MongoDB connection URI.

    db : str
        The name of the database to connect to. Note that if the db doesn't exist on
        the server, no documents will be returned.

    collections : str | list[str]
        The name of the collection(s) to load documents from. If a list is provided,
        documents from all specified collections will be loaded. Note that if the
        collection does not exist on the server, no documents will be returned.
    """

    def __init__(self, uri: str, db: str, collections: str | list[str]):
        self._client = MongoClient(uri)
        self._db = self._client[db]
        self._collections = []
        self._count = None

        # TODO: do we need to check if the collections and database exist first?
        if isinstance(collections, str):
            if collections == "*" or collections == "all":
                self._collections = self._list_all_collections()
            else:
                self._collections.append(self._db[collections])
        else:
            for collection in collections:
                self._collections.append(self._db[collection])

        if len(collections) == 0:
            raise LoaderError("mongodb: no collections specified")

    def __len__(self):
        if self._count is None:
            self._count = sum(
                collection.count_documents({})
                for collection in self._collections
            )
        return self._count

    def __iter__(self):
        self._count = 0
        for collection in self._collections:
            for document in collection.find():
                self._count += 1
                yield document

    def _list_all_collections(self):
        """
        Returns a list of all collections in the database.
        """
        try:
            return self._db.list_collection_names()
        except AttributeError:
            return self._db.collection_names()



"""
Schemaless types for analyzing documents.
"""

import base64

from enum import StrEnum, unique


@unique
class Type(StrEnum):

    AMBIGUOUS = "ambiguous"
    NULL = "null"
    BOOLEAN = "boolean"
    OBJECT = "object"
    ARRAY = "array"
    NUMBER = "number"
    STRING = "string"
    TEXT = "text"
    BLOB = "blob"
    OID = "oid"
    DATETIME = "datetime"
    TIMESTAMP = "timestamp"
    UUID = "uuid"
    ULID = "ulid"


def is_base64(sb: str | bytes) -> bool:
    """
    Check if the value is a correctly encoded base64 encoded string.

    Parameters
    ----------
    value : str
        The value to check.

    Returns
    -------
    bool
        True if the value is a base64 encoded string, False otherwise.
    """
    # Empty strings are not considered base64
    if not sb:
        return False

    try:
        if isinstance(sb, str):
            sb = bytes(sb, "ascii")

        if not isinstance(sb, bytes):
            raise ValueError("argument must be a string or bytes")

        return base64.b64encode(base64.b64decode(sb)) == sb
    except Exception:
        return False



"""
Tests for the configuration module.
"""

import pytest

from schemate.config import BaseConfig
from schemate.exceptions import InvalidConfiguration


class MockConfig(BaseConfig):

    DEFAULTS = {
        "color": "red",
        "size": "large",
    }

    REQUIRED = ["color", "name"]


def test_config_base():
    """
    Test the base functionality of the config object.
    """
    with pytest.raises(InvalidConfiguration):
        _ = MockConfig()

    config = MockConfig({"name": "test", "color": "blue"})
    assert config["name"] == "test"
    assert config["color"] == "blue"
    assert config["size"] == "large"



"""
Defines module and package information for schemate, specifically the version.
"""

# Module version and package information
__version_info__ = {
    "major": 0,
    "minor": 3,
    "micro": 0,
    "releaselevel": "alpha",
    "post": 0,
    "serial": 3,
}


def get_version(short: bool = False) -> str:
    """
    Prints the version.
    """
    assert __version_info__["releaselevel"] in ("alpha", "beta", "final")
    vers = ["{major}.{minor}.{micro}".format(**__version_info__)]

    if __version_info__["releaselevel"] != "final" and not short:
        vers.append(
            "-{}.{}".format(
                __version_info__["releaselevel"],
                __version_info__["serial"],
            )
        )

    if __version_info__["post"]:
        vers.append(".post{}".format(__version_info__["post"]))

    return "".join(vers)



"""
Schemate profile classes for analyzing and generating JSON schema profiles.
"""

import json
import uuid
import datetime
import bson.int64
import bson.binary
import bson.objectid
import bson.timestamp

from .serialize import Encoder
from .types import Type, is_base64
from .exceptions import PropertyValueError, PropertyTypeError

from collections import defaultdict
from dataclasses import dataclass, field
from typing import List, Dict, Union, Any

# If a string is longer than this it is considered a text field.
DEFAULT_TEXT_LIMIT = 256

# If the number of values exceeds this limit then the property is not considered discrete.
DISCRET_VALUES_LIMIT = 50

# Union of all property types that can be inferred from schemaless types.
PropertyType = Union["Property", "ObjectProperty", "ArrayProperty", "AmbiguousProperty"]


@dataclass(init=True, repr=False, eq=True, order=False)
class Profile:
    """
    A profile of a schemaless document or group of documents. The profile contains
    the datatypes of the fields in the documents as well as other information describing
    the fields and their nested properties. The profile can be used to generate a JSON
    schema or another schema format.
    """

    schema: PropertyType
    documents: int = 0
    ambiguous: int = 0

    def dump(self, fp, **kwargs):
        """
        Dumps the profile as a JSON document using the json.dump arguments.
        """
        if "cls" not in kwargs:
            kwargs["cls"] = Encoder
        return json.dump(self, fp, **kwargs)

    def dumps(self, *args, **kwargs):
        """
        Dumps the profile as a JSON string using the json.dumps arguments.
        """
        if "cls" not in kwargs:
            kwargs["cls"] = Encoder
        return json.dumps(self, **kwargs)


def cast(value: Any, text_limit: int = DEFAULT_TEXT_LIMIT) -> PropertyType:
    """
    Cast a value from JSON into a property type. This is the first step in schema
    analysis and is used to determine how to treat the value in the document.
    """
    # Handle standard JSON types
    if value is None:
        return Property(type=Type.NULL, count=1)

    if isinstance(value, bool):
        return Property(type=Type.BOOLEAN, count=1)

    if isinstance(value, int):
        return DiscreteProperty(type=Type.NUMBER, count=1, values={value: 1})

    if isinstance(value, float):
        return Property(type=Type.NUMBER, count=1)

    if isinstance(value, str):
        if len(value) < text_limit:
            return DiscreteProperty(type=Type.STRING, count=1, values={value: 1})
        else:
            if is_base64(value):
                return Property(type=Type.BLOB, count=1)
            return Property(type=Type.TEXT, count=1)

    if isinstance(value, bytes):
        return Property(type=Type.BLOB, count=1)

    if isinstance(value, dict):
        return ObjectProperty(
            type=Type.OBJECT,
            properties={
                k: cast(v) for k, v in value.items()
            },
            count=1,
        )

    if isinstance(value, list):
        # Cast and merge is required to detect the type(s) of the items in the array
        items = [cast(item) for item in value]
        if len(items) == 0:
            return ArrayProperty(type=Type.ARRAY, count=1, items=None)
        if len(items) == 1:
            return ArrayProperty(type=Type.ARRAY, count=1, items=items[0])

        # Merge the items to get the type of the array
        merged = items[0]
        for item in items[1:]:
            merged = merged.merge(item)
        return ArrayProperty(type=Type.ARRAY, count=1, items=merged)

    # The following are BSON or Extended JSON types
    if isinstance(value, datetime.datetime):
        return Property(type=Type.DATETIME, count=1)

    if isinstance(value, bson.objectid.ObjectId):
        return Property(type=Type.OID, count=1)

    if isinstance(value, uuid.UUID):
        return Property(type=Type.UUID, count=1)

    if isinstance(value, bson.binary.Binary):
        return Property(type=Type.BLOB, count=1)

    if isinstance(value, bson.int64.Int64):
        return Property(type=Type.NUMBER, count=1)

    if isinstance(value, bson.timestamp.Timestamp):
        return Property(type=Type.TIMESTAMP, count=1)

    raise PropertyTypeError(
        f"unable to cast value of type {type(value)} to a property type"
    )


@dataclass(init=True, repr=False, eq=False)
class Property:
    """
    A property describes a field in a document as well as the number of times that
    field appears in the document. This is the base class for all more complex
    properties and is used to describe null, boolean, float, text, and blob type
    properties (e.g. properties that can be counted but don't have sub properties or
    discrete values).
    """

    type: Type
    count: int = 0

    def merge(self, other: PropertyType) -> PropertyType:
        """
        Merges the other property into this property. If the other property is not the
        same type as this property then an ambiguous property will be returned.
        """
        # AmbiguousProperty should not call super()
        if self.type == Type.AMBIGUOUS:
            raise PropertyTypeError("unable to merge into ambiguous property")

        # Handle mismatched type ambiguity
        if self.type != other.type:
            if other.type == Type.AMBIGUOUS:
                return other.merge(self)

            return AmbiguousProperty(
                type=Type.AMBIGUOUS,
                types=[self, other],
                count=self.count + other.count,
            )

        # All base classes should update the left-hand side and return self
        self.count += other.count
        return self

    def truncate(self, limit=DISCRET_VALUES_LIMIT) -> "Property":
        return self

    def __eq__(self, other: PropertyType) -> bool:
        if not isinstance(other, Property):
            return False

        return (
            self.type == other.type and self.count == other.count
        )

    def __repr__(self):
        return f"{self.__class__.__name__}{json.dumps(self, cls=Encoder)}"


@dataclass(init=True, repr=False, eq=False, kw_only=True)
class DiscreteProperty(Property):
    """
    A discrete property describes a field that has a fixed set of values such as
    numbers, or strings (note that booleans are not considered discrete). The
    distribution of each value is tracked so long as the
    """

    unique: int = None
    values: Dict[str | int, int] = field(default_factory=defaultdict(int))

    def __post_init__(self):
        # Ensure the values are a defaultdict
        if not isinstance(self.values, defaultdict):
            if isinstance(self.values, dict):
                self.values = defaultdict(int, self.values)
            else:
                raise PropertyTypeError("values must be a dict or a defaultdict")

        # Ignore unique input and calculate it from the values
        self.unique = len(self.values)

    def merge(self, other: PropertyType) -> PropertyType:
        if self.type == other.type:
            if isinstance(other, DiscreteProperty):
                keys = self.values.keys() | other.values.keys()
                for key in keys:
                    self.values[key] += other.values[key]
                self.unique = len(self.values)

            elif isinstance(other, Property):
                return other.merge(self)

            else:
                raise PropertyTypeError(
                    f"unable to merge discrete property with {type(other)}"
                )

        return super(DiscreteProperty, self).merge(other)

    def truncate(self, limit=DISCRET_VALUES_LIMIT) -> Union[Property, "DiscreteProperty"]:
        if len(self.values) > limit:
            return Property(
                type=self.type,
                count=self.count
            )
        return self

    def __eq__(self, other):
        if not isinstance(other, DiscreteProperty):
            return False

        if self.values.keys() != other.values.keys():
            return False

        for key, value in self.values.items():
            if value != other.values[key]:
                return False

        return super(DiscreteProperty, self).__eq__(other)


@dataclass(init=True, repr=False, eq=False, kw_only=True)
class AmbiguousProperty(Property):
    """
    An ambiguous property describes a field that has multiple types. All types observed
    are listed under the types list and the count is the number of times that field
    appeared in a document. The sum of the counts of the types should be equal to the
    top level count of the ambiguous property.
    """

    type: Type = Type.AMBIGUOUS
    types: List[PropertyType] = field(default_factory=list)

    def __post_init__(self):
        self.validate()

    def merge(self, other):
        # Merge both ambiguous types together.
        # Note an ambiguous type cannot contain an ambiguous type.
        if self.type == other.type:
            for item in other.types:
                if item.type == self.type:
                    raise PropertyValueError(
                        "an ambiguous property cannot contain another ambiguous property"
                    )
                self.merge(item)

        # Merge a non-ambiguous type into an ambiguous one.
        else:
            for item in self.types:
                if item.type == other.type:
                    item.merge(other)
                    break
            else:
                self.types.append(other)

            # Cannot call super here to avoid creating another ambiguous property
            self.count += other.count

        # Make sure the merge is validated
        self.validate()
        return self

    def truncate(self, limit=DISCRET_VALUES_LIMIT) -> "AmbiguousProperty":
        self.types = [
            item.truncate(limit) for item in self.types
        ]
        self.validate()
        return self

    def __eq__(self, other):
        # Only valid types can be compared.
        self.validate()

        if not isinstance(other, AmbiguousProperty):
            return False

        # Ensure the other ambiguous property is valid
        other.validate()

        # Order doesn't matter for equality
        if len(self.types) != len(other.types):
            return False

        for item in self.types:
            for cmpt in other.types:
                if item.type == cmpt.type:
                    if item != cmpt:
                        return False
                    break
            else:
                return False

        return super(AmbiguousProperty, self).__eq__(other)

    def validate(self):
        types = [t.type for t in self.types]
        if len(self.types) != len(set(types)):
            raise PropertyValueError(
                "ambiguous property contains duplicate types"
            )

        if Type.AMBIGUOUS in types:
            raise PropertyValueError(
                "an ambiguous property cannot contain another ambiguous property"
            )


@dataclass(init=True, repr=False, eq=False, kw_only=True)
class ObjectProperty(Property):
    """
    Allows nested properties to be defined for a document or for fields in the document.
    The properties are defined as a dictionary of field names to their properties. The
    count is the number of times the field appeared in a document.
    """

    type: Type = Type.OBJECT
    properties: Dict[str, PropertyType] = field(default_factory=dict)

    def merge(self, other):
        if self.type == other.type:
            # Merge the properties of the other object into this one
            for key, prop in other.properties.items():
                if key in self.properties:
                    self.properties[key].merge(prop)
                else:
                    self.properties[key] = prop

        return super(ObjectProperty, self).merge(other)

    def truncate(self, limit=DISCRET_VALUES_LIMIT) -> "ObjectProperty":
        for key, prop in self.properties.items():
            self.properties[key] = prop.truncate(limit)
        return self

    def __eq__(self, other):
        if not isinstance(other, ObjectProperty):
            return False

        if self.properties.keys() != other.properties.keys():
            return False

        for key, value in self.properties.items():
            if value != other.properties[key]:
                return False

        return super(ObjectProperty, self).__eq__(other)


@dataclass(init=True, repr=False, eq=False, kw_only=True)
class ArrayProperty(Property):
    """
    Allows for a property to be a list and define what the properties are of each of
    the items in the list. If the array has multiple types then it will be an
    AmbiguousProperty that describes each of the items in the array.
    """

    type: Type = Type.ARRAY
    items: PropertyType

    def merge(self, other):
        if self.type == other.type:
            if self.items is None:
                self.items = other.items
            elif other.items is not None:
                self.items.merge(other.items)
        return super(ArrayProperty, self).merge(other)

    def truncate(self, limit=DISCRET_VALUES_LIMIT) -> "ArrayProperty":
        if self.items is not None:
            self.items = self.items.truncate(limit)
        return self

    def __eq__(self, other):
        if not isinstance(other, ArrayProperty):
            return False

        if self.items != other.items:
            return False

        return super(ArrayProperty, self).__eq__(other)



"""
Test the schemaless types module.
"""

import pytest

from schemate.types import *


@pytest.mark.parametrize(
    "sb,expected",
    [
        ("", False),
        ("aGVsbG8gd29ybGQ=", True),
        ("gqSb/4Kkm6d+pJvLgqSb/4Kkl/+CpJf/fqSbl36kmJ9+pJgA", True),
        ("not base 64", False),
        ("TG9uZyBiYXNlNjQgZW5jb2RlZCBzdHJpbmcsIGhvcGVmdWxseS4", False),
        ("TG9uZyBiYXNlNjQgZW5jb2RlZCBzdHJpbmcsIGhvcGVmdWxseS4=", True),
        ("TG9uZyBiYXNlNjQgZW5jb2RlZCBzdHJpbmcsIGhvcGVmdWxseSAt", True),
        ("TG9uZyBiYXNlNjQgZW5jb2RlZCBzdHJpbmcsIGhvcGVmdWxseSAtLQ=", False),
        ("TG9uZyBiYXNlNjQgZW5jb2RlZCBzdHJpbmcsIGhvcGVmdWxseSAtLQ", False),
        ("TG9uZyBiYXNlNjQgZW5jb2RlZCBzdHJpbmcsIGhvcGVmdWxseSAtLQ==", True),
        (None, False),
        (42, False)
    ],
)
def test_is_base64(sb, expected):
    """
    Test the is_base64 function.
    """
    assert is_base64(sb) is expected



"""
Utility to analyze the JSON schema of a document or a group of documents.
"""

from .types import Type
from .loaders import Loader
from .schemate import Profile, PropertyType, cast


class SchemaAnalysis(object):
    """
    Loads schemaless documents and analyzes the keys, their values, and their
    properties. Returns a schema profile of the documents that can be saved to disk.

    Parameters
    ----------
    loader : Loader
        A loader that provides the documents to be analyzed. The loader must
        implement the `__iter__` method to yield documents.
    """

    def __init__(self, loader: Loader):
        self._loader = loader
        self._result = None

    @property
    def result(self):
        return self._result

    def run(self):
        """
        Perform an analysis on the documents provided by the loader.
        """
        self._result = Profile(schema=None)
        for document in self._loader:
            self.analyze(document)

        self._result.schema.truncate()
        self._result.ambiguous = self.ambiguous(self._result.schema)

    def analyze(self, document):
        # Count the number of documents in the dataset
        self._result.documents += 1

        # Update the schema
        if self._result.schema is None:
            self._result.schema = cast(document)
        else:
            self._result.schema.merge(cast(document))

    def ambiguous(self, property: PropertyType) -> int:
        if property.type == Type.AMBIGUOUS:
            return 1 + sum(self.ambiguous(p) for p in property.types)
        elif property.type == Type.OBJECT:
            return sum(self.ambiguous(p) for p in property.properties.values())
        elif property.type == Type.ARRAY and property.items is not None:
            return self.ambiguous(property.items)
        else:
            return 0



"""
Tests the schemate property analysis classes.
"""

import json
import pytest
import base64

from io import StringIO
from schemate.schemate import *
from schemate.exceptions import *
from collections import namedtuple


def test_profile_dump():
    """
    Test profile serialization to JSON format.
    """
    profile = Profile(
        schema=ObjectProperty(properties={
            "available": Property(type="boolean", count=10),
            "color": DiscreteProperty(type="string", count=10, values={"red": 5, "blue": 5}),
            "size": DiscreteProperty(type="number", count=10, values={1: 3, 2: 3, 3: 1}),
            "names": ArrayProperty(
                items=Property(type="string", count=10),
                count=7,
            ),
            "tags": AmbiguousProperty(
                types=[
                    Property(type="string", count=5),
                    ObjectProperty(
                        properties={
                            "key": Property(type="string", count=5),
                            "value": Property(type="string", count=5),
                        },
                        count=5,
                    ),
                ],
                count=10,
            ),
        }, count=10),
        documents=10,
        ambiguous=1,
    )

    expected = {
        "schema": {
            "type": "object",
            "count": 10,
            "properties": {
                "available": {
                    "type": "boolean",
                    "count": 10,
                },
                "color": {
                    "type": "string",
                    "count": 10,
                    "unique": 2,
                    "values": {
                        "red": 5,
                        "blue": 5,
                    },
                },
                "size": {
                    "type": "number",
                    "count": 10,
                    "unique": 3,
                    "values": {
                        "1": 3,
                        "2": 3,
                        "3": 1,
                    },
                },
                "names": {
                    "type": "array",
                    "count": 7,
                    "items": {
                        "type": "string",
                        "count": 10,
                    },
                },
                "tags": {
                    "type": "ambiguous",
                    "count": 10,
                    "types": [
                        {
                            "type": "string",
                            "count": 5,
                        },
                        {
                            "type": "object",
                            "properties": {
                                "key": {
                                    "type": "string",
                                    "count": 5,
                                },
                                "value": {
                                    "type": "string",
                                    "count": 5,
                                },
                            },
                            "count": 5,
                        },
                    ],
                },
            },
        },
        "documents": 10,
        "ambiguous": 1,
    }

    result = profile.dumps()
    assert json.loads(result) == expected

    fp = StringIO()
    profile.dump(fp)
    fp.seek(0)
    result = json.load(fp)
    assert result == expected


@pytest.mark.parametrize(
    "value,expected",
    [
        (None, Property(type="null", count=1)),
        (True, Property(type="boolean", count=1)),
        (False, Property(type="boolean", count=1)),
        (1, DiscreteProperty(type="number", count=1, values={1: 1})),
        (2.0, Property(type="number", count=1)),
        ("hello", DiscreteProperty(type="string", count=1, values={"hello": 1})),
        (
            "This is a very long piece of text that is over two hundred and fifty five characters long. It has more text inside of it than you can shake a stick at. A milkshake would be delicious right now but instead I have to keep typing this extra long string. It's long enough now.",
            Property(type="text", count=1),
        ),
        (
            base64.b64encode(b"This is a very long piece of text that is over two hundred and fifty five characters long. It has more text inside of it than you can shake a stick at. A milkshake would be delicious right now but instead I have to keep typing this extra long string. It's long enough now.").decode('ascii'),
            Property(type="blob", count=1),
        ),
        (
            base64.b64decode(
                "TG9uZyBiYXNlNjQgZW5jb2RlZCBzdHJpbmcsIGhvcGVmdWxseSAtLQ=="
            ),
            Property(type="blob", count=1),
        ),
        (
            {"key": "value"},
            ObjectProperty(
                type="object",
                properties={
                    "key": DiscreteProperty(
                        type="string", count=1, values={"value": 1}
                    ),
                },
                count=1,
            ),
        ),
        (
            ["red", "green", "blue", "red"],
            ArrayProperty(
                items=DiscreteProperty(
                    type="string",
                    count=4,
                    unique=3,
                    values={"red": 2, "green": 1, "blue": 1},
                ),
                count=1,
            ),
        ),
        ([], ArrayProperty(items=None, count=1)),
        ([13.1321], ArrayProperty(items=Property(type="number", count=1), count=1)),
    ],
)
def test_cast(value, expected):
    """
    Test casting a value to a property type.
    """
    assert cast(value) == expected


@pytest.mark.parametrize(
    "alpha,bravo,expected",
    [
        (cast(None), cast(None), Property(type="null", count=2)),
        (cast(True), cast(False), Property(type="boolean", count=2)),
        (
            cast(1),
            cast(1),
            DiscreteProperty(type="number", count=2, values={1: 2}),
        ),
        (
            cast(1),
            cast(2),
            DiscreteProperty(type="number", count=2, values={1: 1, 2: 1}),
        ),
        (cast(3.14), cast(1.221), Property(type="number", count=2)),
        (
            cast("hello"),
            cast("world"),
            DiscreteProperty(type="string", count=2, values={"hello": 1, "world": 1}),
        ),
        (
            cast("foo"),
            cast("foo"),
            DiscreteProperty(type="string", count=2, values={"foo": 2}),
        ),
        (
            DiscreteProperty(type="string", count=21, values={"red": 13, "green": 8}),
            DiscreteProperty(type="string", count=6, values={"red": 2, "blue": 4}),
            DiscreteProperty(
                type="string",
                count=27,
                values={"red": 15, "green": 8, "blue": 4},
            ),
        ),
        (
            cast({"size": 14, "color": "red"}),
            cast({"size": 14, "color": "blue"}),
            ObjectProperty(
                type="object",
                properties={
                    "size": DiscreteProperty(type="number", count=2, values={14: 2}),
                    "color": DiscreteProperty(
                        type="string", count=2, values={"red": 1, "blue": 1}
                    ),
                },
                count=2,
            ),
        ),
        (
            cast({"size": 14, "color": "red", "tags": ["foo", "bar"]}),
            cast({"size": 14, "color": "blue", "name": "Shirt"}),
            ObjectProperty(
                type="object",
                properties={
                    "name": DiscreteProperty(
                        type="string", count=1, unique=1, values={"Shirt": 1}
                    ),
                    "size": DiscreteProperty(
                        type="number", count=2, unique=1, values={14: 2}
                    ),
                    "color": DiscreteProperty(
                        type="string", count=2, unique=2, values={"red": 1, "blue": 1}
                    ),
                    "tags": ArrayProperty(
                        items=DiscreteProperty(
                            type="string",
                            count=2,
                            unique=2,
                            values={"foo": 1, "bar": 1},
                        ),
                        count=1,
                    ),
                },
                count=2,
            ),
        ),
        (
            cast(["red", "green", "blue"]),
            cast(["red", "green", "yellow"]),
            ArrayProperty(
                items=DiscreteProperty(
                    type="string",
                    count=6,
                    unique=5,
                    values={"red": 2, "green": 2, "blue": 1, "yellow": 1},
                ),
                count=2,
            ),
        ),
        (
            cast(3.14),
            AmbiguousProperty(
                types=[
                    Property(type="number", count=1),
                    Property(type="boolean", count=1),
                ],
                count=2,
            ),
            AmbiguousProperty(
                types=[
                    Property(type="boolean", count=1),
                    Property(type="number", count=2),
                ],
                count=3,
            ),
        ),
        (
            cast(3.14),
            AmbiguousProperty(
                types=[
                    Property(type="null", count=1),
                    Property(type="boolean", count=1),
                ],
                count=2,
            ),
            AmbiguousProperty(
                types=[
                    Property(type="boolean", count=1),
                    Property(type="null", count=1),
                    Property(type="number", count=1),
                ],
                count=3,
            ),
        ),
        (
            AmbiguousProperty(
                types=[
                    Property(type="null", count=1),
                    Property(type="boolean", count=1),
                ],
                count=2,
            ),
            AmbiguousProperty(
                types=[
                    Property(type="null", count=2),
                    Property(type="boolean", count=1),
                ],
                count=2,
            ),
            AmbiguousProperty(
                types=[
                    Property(type="boolean", count=2),
                    Property(type="null", count=3),
                ],
                count=5,
            )
        ),
        (
            AmbiguousProperty(
                types=[
                    Property(type="null", count=1),
                    Property(type="text", count=8),
                ],
                count=9,
            ),
            AmbiguousProperty(
                types=[
                    Property(type="null", count=3),
                    Property(type="boolean", count=4),
                ],
                count=7,
            ),
            AmbiguousProperty(
                types=[
                    Property(type="boolean", count=4),
                    Property(type="null", count=4),
                    Property(type="text", count=8),
                ],
                count=16,
            ),
        ),
    ],
)
def test_property_merge(alpha, bravo, expected):
    """
    Test merging properties.
    """
    assert alpha.merge(bravo) == expected


def test_discrete_values():
    """
    Assert discrete values must be a defaultdict or dict.
    """
    with pytest.raises(PropertyTypeError):
        DiscreteProperty(
            type="string",
            count=10,
            values=42,
        )


def test_no_nested_ambiguous():
    """
    Assert ambiguous types cannot contain other ambiguous types.
    """
    with pytest.raises(PropertyValueError):
        AmbiguousProperty(
            types=[
                Property(type="string", count=5),
                AmbiguousProperty(
                    types=[
                        Property(type="string", count=5),
                        Property(type="number", count=5),
                    ],
                    count=5,
                ),
            ],
            count=10,
        )

    with pytest.raises(PropertyValueError):
        prop = AmbiguousProperty(
            types=[
                Property(type="string", count=10),
            ],
            count=10,
        )

        MockAmbiguousProperty = namedtuple(
            "MockAmbiguousProperty", ["type", "types", "count"]
        )

        mock = MockAmbiguousProperty(
            type="ambiguous",
            types=[
                Property(type="string", count=5),
                AmbiguousProperty(types=[Property(type="number", count=5)], count=5),
            ],
            count=5,
        )

        prop.merge(mock)



"""
Schemate exception hierarchy
"""


class SchemateException(Exception):
    """Base class for all Schemate exceptions."""
    pass


class CommandError(SchemateException):
    """CLI errors that are printed without a traceback."""
    pass


class ConfigLoadError(SchemateException, ValueError):
    pass


class InvalidConfiguration(SchemateException):
    pass


class LoaderError(SchemateException):
    """Errors related to data loaders."""
    pass


class UnsupportedLoader(LoaderError, ValueError):
    pass


class PropertyTypeError(SchemateException, TypeError):
    pass


class PropertyValueError(SchemateException, ValueError):
    pass



"""
Test the document loaders.
"""

import pytest

from schemate.loaders import *


def test_loader_base_class():

    class MockLoader(Loader):

        def __iter__(self):
            yield {"color": "red", "size": "large"}
            yield {"color": "blue", "size": "small"}
            yield {"color": "red"}

        def __len__(self):
            return 3

    loader = MockLoader()
    assert loader.count() == 3


@pytest.mark.parametrize(
    "path, expected",
    [
        ("/path/to/file.json", True),
        ("/path/to/file.JSON", True),
        ("/path/to/file.Json", True),
        ("file.json", True),
        ("/path/to/file.jsonl", True),
        ("/path/to/file.jsonlines", True),
        ("/path/to/file.yaml", True),
        ("/path/to/file.yml", True),
        ("/path/to/file.txt", False),
        ("file.txt", False),
        ("file", False),
        ("", False),
    ],
)
def test_file_loader_is_supported(path, expected):
    """
    Test the file loader's is_supported method.
    """
    assert FileLoader.is_supported(path) is expected



"""
Helpers for serializing and deserializing python types from JSON
"""

import json
import enum
import dataclasses


class Encoder(json.JSONEncoder):

    def default(self, o):
        if dataclasses.is_dataclass(o):
            return dataclasses.asdict(o)

        if isinstance(o, enum.Enum):
            return o.value

        return super(Encoder, self).default(o)



"""
Configure an analysis or schema generation job.
"""

import os
import abc
import copy
import json
import yaml
import inspect

from typing import Any
from collections.abc import Mapping

from .exceptions import ConfigLoadError, InvalidConfiguration


class BaseConfig(Mapping):
    """
    Base configuration object and ABC to define configuration handling.
    """

    @classmethod
    @abc.abstractmethod
    def DEFAULTS(cls) -> dict:
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def REQUIRED(cls) -> list:
        raise NotImplementedError

    def __init__(self, data=None):
        self._config = copy.deepcopy(self.DEFAULTS)
        for key, value in self._config.items():
            if inspect.isclass(value):
                self._config[key] = value()

        if data:
            self.update(data)
        self.validate()

    def update(self, conf: dict):
        """
        Update the configuration object with new values.
        NOTE: this is the only way to mutate the configuration object.
        """
        for key, value in conf.items():
            if key in self._config:
                orig = self[key]
                if isinstance(orig, dict):
                    orig.update(value)
                elif hasattr(orig, "update"):
                    orig.update(value)
                else:
                    self._config[key] = value
            else:
                self._config[key] = value

    def validate(self):
        """
        Check that any required configuration is present.
        """
        for key in self.REQUIRED:
            if key not in self or not self[key]:
                raise InvalidConfiguration(f"missing required configuration: {key}")

            if hasattr(self[key], "validate"):
                self[key].validate()

    def __getitem__(self, key: str) -> Any:
        return self._config[key]

    def __iter__(self):
        return iter(self._config)

    def __len__(self) -> int:
        return len(self._config)


class LoaderConfig(BaseConfig):

    DEFAULTS = {"type": None}
    REQUIRED = ["type"]
    TYPES = {
        "glob": "schemate.loaders.GlobLoader",
        "mongo": "schemate.loaders.MongoDBLoader",
        "mongodb": "schemate.loaders.MongoDBLoader",
        "directory": "schemate.loaders.DirectoryLoader",
        "dir": "schemate.loaders.DirectoryLoader",
        "jsonlines": "schemate.loaders.JSONLinesLoader",
        "jsonl": "schemate.loaders.JSONLinesLoader",
        "files": "schemate.loaders.FilesLoader",
    }

    def validate(self):
        super(LoaderConfig, self).validate()
        loader = self["type"].lower().strip()
        if loader not in self.TYPES:
            raise InvalidConfiguration(
                f"invalid loader type: {loader}"
            )


class AnalysisConfig(BaseConfig):

    DEFAULTS = {"loader": LoaderConfig}
    REQUIRED = ["loaders"]


class Config(BaseConfig):

    DEFAULTS = {
        "analyze": AnalysisConfig,
    }

    REQUIRED = [
        "analyze",
    ]

    @classmethod
    def load(cls, path: str) -> "Config":
        """
        Load a configuration file.
        """
        loader = None
        _, ext = os.path.splitext(os.path.basename(path))

        if ext == ".json":
            loader = json.load
        elif ext == ".yaml" or ext == ".yml":
            loader = yaml.safe_load
        else:
            raise ConfigLoadError(
                f"unsupported config file type: {ext}: supported types are .json and .yaml"
            )

        try:
            with open(path, "r") as f:
                data = loader(f)
                return cls(data)
        except FileNotFoundError as e:
            raise ConfigLoadError(f"config file not found: {path}") from e
        except Exception as e:
            raise ConfigLoadError("failed to load config file") from e



"""
Setup script for schemate
"""

import os
import codecs

from setuptools import setup
from setuptools import find_packages


## Basic Information
NAME = "schemate"
DESCRIPTION = "Analyzes NoSQL or JSON documents to try to infer a schema or general properties."  # noqa
AUTHOR = "Rotational Labs"
EMAIL = "support@rotational.io"
MAINTAINER = "Rotational Labs"
LICENSE = "BSD 3"
REPOSITORY = "https://github.com/rotationalio/llm-benchmark"
PACKAGE = "schemate"


## Define the keywords
KEYWORDS = [
    "json",
    "nosql",
    "schema",
]


## Define the classifiers
## See https://pypi.python.org/pypi?%3Aaction=list_classifiers
CLASSIFIERS = [
    "Development Status :: 3 - Alpha",
    "Intended Audience :: Developers",
    "Intended Audience :: System Administrators",
    "License :: OSI Approved :: BSD License",
    "Natural Language :: English",
    "Operating System :: OS Independent",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3.12",
    "Topic :: System :: Benchmark",
    "Topic :: System :: Hardware",
]


## Important Paths
PROJECT = os.path.abspath(os.path.dirname(__file__))
REQUIRE_PATH = "requirements.txt"
VERSION_PATH = os.path.join(PACKAGE, "version.py")
PKG_DESCRIBE = "README.md"


## Directories to ignore in find_packages
EXCLUDES = [
    "tests",
    "tests.*",
    "bin",
    "docs",
    "docs.*",
    "fixtures",
]


def read(*parts):
    """
    Assume UTF-8 encoding and return the contents of the file located at the
    absolute path from the REPOSITORY joined with *parts.
    """
    with codecs.open(os.path.join(PROJECT, *parts), "rb", "utf-8") as f:
        return f.read()


def get_version(path=VERSION_PATH):
    """
    Reads the python file defined in the VERSION_PATH to find the get_version
    function, and executes it to ensure that it is loaded correctly. Separating
    the version in this way ensures no additional code is executed.
    """
    namespace = {}
    exec(read(path), namespace)
    return namespace["get_version"](short=True)


def get_requires(path=REQUIRE_PATH):
    """
    Yields a generator of requirements as defined by the REQUIRE_PATH which
    should point to a requirements.txt output by `pip freeze`.
    """
    for line in read(path).splitlines():
        line = line.strip()
        if line and not line.startswith("#"):
            yield line


def get_description_type(path=PKG_DESCRIBE):
    """
    Returns the long_description_content_type based on the extension of the
    package describe path (e.g. .txt, .rst, or .md).
    """
    _, ext = os.path.splitext(path)
    return {".rst": "text/x-rst", ".txt": "text/plain", ".md": "text/markdown"}[ext]


if __name__ == "__main__":
    config = {
        "name": NAME,
        "version": get_version(),
        "description": DESCRIPTION,
        "long_description": read(PKG_DESCRIBE),
        "long_description_content_type": get_description_type(PKG_DESCRIBE),
        "classifiers": CLASSIFIERS,
        "keywords": KEYWORDS,
        "license": LICENSE,
        "author": AUTHOR,
        "author_email": EMAIL,
        "maintainer": MAINTAINER,
        "maintainer_email": EMAIL,
        "project_urls": {
            "Download": "{}/tarball/v{}".format(REPOSITORY, get_version()),
            "Source": REPOSITORY,
            "Tracker": "{}/issues".format(REPOSITORY),
        },
        "download_url": "{}/tarball/v{}".format(REPOSITORY, get_version()),
        "packages": find_packages(where=PROJECT, exclude=EXCLUDES),
        "package_data": {},
        "zip_safe": True,
        "entry_points": {
            "console_scripts": [
                "schemate = schemate.__main__:main",
            ]
        },
        "install_requires": list(get_requires()),
        "python_requires": ">=3.10, <4",
    }

    setup(**config)
